import functools
import contextvars
import grpc
from typing import Callable, List, get_type_hints, Type, NamedTuple, Dict, Iterable

from google.protobuf.message import Message

from .grpc_services import load_all_grpc_services, GRPCMethod, GRPCService
from .exceptions import MultipleServicesDefined, NoServicesDefined, AmbiguousMethod, IncorrectMethodArguments, \
    AmbiguousGRPCMethod
from . import log

context = contextvars.ContextVar('petal.context')

__all__ = ['Service']


class Handler(NamedTuple):
    name: str
    function: Callable
    input: Type
    output: Type

    @property
    def python_name(self):
        return self.function.__name__

    def create_method_handler(self) -> grpc.RpcMethodHandler:
        stream_input = hasattr(self.input, '__origin__') and issubclass(self.input.__origin__, Iterable)
        stream_output = hasattr(self.output, '__origin__') and issubclass(self.output.__origin__, Iterable)
        input_type = self.input if not stream_input else self.input.__args__[0]
        output_type = self.output if not stream_output else self.output.__args__[0]

        handlers = {
            (False, False): grpc.unary_unary_rpc_method_handler,
            (True, False): grpc.stream_unary_rpc_method_handler,
            (False, True): grpc.unary_stream_rpc_method_handler,
            (True, True): grpc.stream_stream_rpc_method_handler,
        }

        constructor = handlers[(stream_input, stream_output)]
        return constructor(
            self.function,
            request_deserializer=input_type.FromString,
            response_serializer=output_type.SerializeToString
        )


def camel_case_name(name):
    return ''.join(p.capitalize() for p in name.split('_'))


@functools.lru_cache()
def load_grpc_service(package, service_name=None) -> GRPCService:
    all_services = load_all_grpc_services(package)

    if not all_services:
        raise NoServicesDefined()

    if len(all_services) > 1:
        if not service_name:
            raise MultipleServicesDefined(all_services.keys())
        service = all_services[service_name]
    else:
        service = all_services[list(all_services.keys())[0]]

    return service


class Service:
    def __init__(self, package: str, service_name: str = None):
        self.package = package
        self.service_name = service_name
        self.rpc_methods: List[Handler] = []
        self.logger = log.logger

    def dispatch(self, request_object: Message, context_object: grpc.ServicerContext, func: Callable):
        ctx = contextvars.copy_context()
        context.set(context_object)

        bound_logger = self.logger.bind(peer=context_object.peer())

        with bound_logger.catch():
            try:
                return ctx.run(func, request_object)
            except exceptions.GRPCError as e:
                context_object.set_code(e.code)
                if e.details:
                    context_object.set_details(e.details)
            except NotImplementedError:
                context_object.set_code(grpc.StatusCode.UNIMPLEMENTED)

    def grpc(self, name: str = None):
        def inner(func):
            @functools.wraps(func)
            def wrapper(request_object, context_object):
                return self.dispatch(request_object, context_object, func)

            hints = get_type_hints(func)
            request_type = hints['request']
            return_type = hints['return']

            self.rpc_methods.append(
                Handler(
                    name=name or camel_case_name(func.__name__),
                    function=wrapper,
                    input=request_type,
                    output=return_type,
                )
            )

            return wrapper

        return inner

    def create_service_handler(self) -> 'grpc.ServiceRpcHandler':
        handlers = {
            method.name: method.create_method_handler()
            for method in self.rpc_methods
        }
        service = load_grpc_service(self.package, self.service_name)
        return grpc.method_handlers_generic_handler(
            service.full_name,
            handlers
        )

    def validate_service(self):
        service = load_grpc_service(self.package, self.service_name)
        method_name_mapping: Dict[str, GRPCMethod] = {method.name: method for method in service.methods}
        found_methods = set()

        for method in self.rpc_methods:
            if method.name not in method_name_mapping:
                raise AmbiguousMethod(method.name)
            grpc_method = method_name_mapping[method.name]
            # Check the types are correct
            input_type = grpc_method.input_type
            output_type = grpc_method.output_type

            if grpc_method.input_stream:
                input_type = Iterable[input_type]
            if grpc_method.output_stream:
                output_type = Iterable[output_type]

            if method.input != input_type:
                raise IncorrectMethodArguments(method.python_name, 'input', method.input, input_type)
            if method.output != output_type:
                raise IncorrectMethodArguments(method.python_name, 'output', method.output, output_type)
            found_methods.add(grpc_method)

        missing_methods = set(service.methods) - found_methods
        if missing_methods:
            raise AmbiguousGRPCMethod(name=list(missing_methods)[0].name)



import pkgutil

import importlib
from typing import NamedTuple, List, Dict, Generator, Type
from types import ModuleType
from google.protobuf import symbol_database

pb_symbol_database = symbol_database.Default()


class GRPCMethod(NamedTuple):
    name: str
    full_name: str
    input_type: Type
    input_stream: bool
    output_type: Type
    output_stream: bool


class GRPCService(NamedTuple):
    name: str
    full_name: str
    methods: List[GRPCMethod]


def load_all_grpc_services(package: str) -> Dict[str, GRPCService]:
    services = {}

    for module in iter_modules(package):
        for service_name, service_descriptor in module.DESCRIPTOR.services_by_name.items():

            service = GRPCService(name=service_name,
                                  full_name=service_descriptor.full_name,
                                  methods=[])
            for method_descriptor in service_descriptor.methods:
                stream_info = module.STREAMING_INFO[service_name][method_descriptor.name]
                input_type = pb_symbol_database.GetSymbol(method_descriptor.input_type.full_name)
                output_type = pb_symbol_database.GetSymbol(method_descriptor.output_type.full_name)
                method = GRPCMethod(name=method_descriptor.name,
                                    full_name=method_descriptor.full_name,
                                    input_type=input_type,
                                    output_type=output_type,
                                    input_stream=stream_info['client_streaming'],
                                    output_stream=stream_info['server_streaming']
                                    )
                service.methods.append(method)
            services[service.name] = service

    return services


def iter_modules(package: str) -> Generator[ModuleType, None, None]:
    for module_info in pkgutil.iter_modules([f'{package}/protobuf']):
        if not module_info.name.endswith('_pb2'):
            continue
        yield importlib.import_module(f'{package}.protobuf.{module_info.name}')



from typing import Iterable, Any

from grpc import StatusCode


class InitializationException(Exception):
    pass


class MultipleServicesDefined(InitializationException):
    def __init__(self, service_names: Iterable[str]):
        self.service_names = service_names

    def __str__(self):
        return f'Multiple GRPC services are defined. ' \
            f'Please use Service(..., service_name=NAME) to specify one.\n' \
            f'Services found: {", ".join(self.service_names)}.'


class NoServicesDefined(InitializationException):
    def __str__(self):
        return 'No services are defined. ' \
               'Please ensure one exists in your protobuf file.'


class AmbiguousMethod(InitializationException):
    def __init__(self, name: str):
        self.name = name

    def __str__(self):
        return f'Cannot find a suitable GRPC method for function {self.name}\n' \
            f'Please ensure one exists in your .proto file, or use service.grpc(name=NAME) ' \
            f'to define one.'


class AmbiguousGRPCMethod(InitializationException):
    def __init__(self, name: str):
        self.name = name

    def __str__(self):
        return f'Cannot find a suitable method for GRPC method {self.name}\n' \
            f'Please ensure one exists in your service code, or use service.grpc(name=NAME) ' \
            f'to define one.'



class IncorrectMethodArguments(InitializationException):
    def __init__(self, method_name: str, argument: str, given: Any, expected: Any):
        self.method_name = method_name
        self.argument = argument
        self.given = given
        self.expected = expected

    def __str__(self):
        return f'Method {self.method_name} has an incorrect {self.argument} \n' \
            f'Expected {self.expected}, found {self.given}.\n' \
            f'Please update your service definitions to ensure the types are correct.'


class GRPCError(Exception):
    code = None
    details = ""

    def __init__(self, details=None, code=None):
        if code:
            self.code = code
        if details:
            self.details = details


class Cancelled(GRPCError):
    code = StatusCode.CANCELLED


class InvalidArgument(GRPCError):
    code = StatusCode.INVALID_ARGUMENT


class DeadlineExceeded(GRPCError):
    code = StatusCode.DEADLINE_EXCEEDED


class NotFound(GRPCError):
    code = StatusCode.NOT_FOUND


class AlreadyExists(GRPCError):
    code = StatusCode.NOT_FOUND


class PermissionDenied(GRPCError):
    code = StatusCode.PERMISSION_DENIED


class Unauthenticated(GRPCError):
    code = StatusCode.UNAUTHENTICATED


class ResourceExhausted(GRPCError):
    code = StatusCode.RESOURCE_EXHAUSTED


class FailedPrecondition(GRPCError):
    code = StatusCode.FAILED_PRECONDITION


class Aborted(GRPCError):
    code = StatusCode.ABORTED


class Unimplemented(GRPCError):
    code = StatusCode.UNIMPLEMENTED


class InternalError(GRPCError):
    code = StatusCode.INTERNAL


class Unavailable(GRPCError):
    code = StatusCode.UNAVAILABLE


class DataLoss(GRPCError):
    code = StatusCode.DATA_LOSS



import time
from concurrent import futures

import click
import os
from pathlib import Path
import importlib
import sys
import subprocess

import grpc
import pkg_resources

from petal import Service
from petal.log import logger
from petal.exceptions import InitializationException


@click.group()
def cli():
    pass


@cli.command()
@click.argument('module')
@click.option('--bind', default='0.0.0.0:50051')
@click.option('--shutdown-grace', default=10)
@click.option('--threads', default=10, type=int)
def run(module, bind, shutdown_grace, threads):
    sys.path.append(os.getcwd())
    module_object = importlib.import_module(module)

    try:
        app: Service = module_object.service
    except AttributeError:
        raise click.UsageError(f'Could not find {module}.service')

    try:
        app.validate_service()
    except InitializationException as e:
        raise click.ClickException(str(e)) from e

    handler = app.create_service_handler()
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=threads))
    server.add_generic_rpc_handlers((handler,))
    server.add_insecure_port(bind)
    server.start()
    logger.info(f'Started Petal. Listening on {bind}')
    try:
        while True:
            time.sleep(60 * 60 * 24)
    except KeyboardInterrupt:
        logger.info(f'Stopping. Closing all connections after {shutdown_grace} seconds.')
        server.stop(grace=10)
        logger.info('Successfully stopped.')


@cli.command()
@click.argument('service_directory', type=click.Path(exists=True, dir_okay=True, file_okay=False))
def build(service_directory):
    service_directory = Path(service_directory)
    proto_directory = service_directory / 'protobuf'
    if not proto_directory.exists():
        raise click.UsageError(f'{proto_directory} does not exist.')

    click.echo('Clearing output directory...')
    proto_files = set(proto_directory.glob('*.proto')) | {proto_directory / '__init__.py'}
    for path in proto_directory.iterdir():
        if path.is_file() and path not in proto_files:
            path.unlink()

    includes = []
    for entry_point in pkg_resources.iter_entry_points('petal_include_protobuf'):
        entry_point = Path(entry_point.load().__file__)
        if not entry_point.exists():
            logger.warning(f'Registered include path {entry_point} does not exist')
            continue

        includes.append(f'-I={entry_point.parent}')

    proto_files = [str(f.relative_to(service_directory.parent)) for f in proto_directory.glob('*.proto')]

    args = [
        'python3',
        '-m',
        'grpc_tools.protoc',
        f'--proto_path={service_directory.parent}/',
        f'--grpc_python_out={service_directory.parent}/',
        f'--python_out={service_directory.parent}/',
        f'--extract-streaming_out={service_directory.parent}/',
        f'--mypy_out={service_directory.parent}/',
        *includes,
        *proto_files
    ]
    click.echo('Compiling protobufs with ', nl=False)
    click.secho(f'{" ".join(args)}', fg='green')
    subprocess.check_call(args)


def main():
    cli()



import sys
import json

from google.protobuf.compiler import plugin_pb2
from google.protobuf.json_format import MessageToDict


def generate_service_metadata(request: plugin_pb2.CodeGeneratorRequest,
                              response: plugin_pb2.CodeGeneratorResponse):
    files = {f.name: f for f in request.proto_file}
    to_generate = {n: files[n] for n in request.file_to_generate}

    for name, fd in to_generate.items():

        services = {
            service.name: {
                method.name: {
                    'client_streaming': method.client_streaming,
                    'server_streaming': method.server_streaming
                }
                for method in service.method
            }
            for service in fd.service
        }

        if not services:
            continue

        output = response.file.add()
        output.insertion_point = 'module_scope'
        output.name = fd.name[:-6].replace('-', '_') + '_pb2.py'
        output.content = 'STREAMING_INFO = %s' % services
        print(f'Writing structure to {output.name}', file=sys.stderr)


def main():
    data = sys.stdin.buffer.read()
    request = plugin_pb2.CodeGeneratorRequest()
    request.ParseFromString(data)

    response = plugin_pb2.CodeGeneratorResponse()
    generate_service_metadata(request, response)
    output = response.SerializeToString()

    sys.stdout.buffer.write(output)


if __name__ == '__main__':
    main()



from setuptools import setup

# read the contents of your README file
from pathlib import Path

readme = Path(__file__).parent / 'README.md'
if readme.exists():
    with readme.open('r', encoding='utf-8') as fd:
        long_description = fd.read()
else:
    long_description = ''

setup(
    name='petal',
    version='0.0.2-1',
    packages=['petal'],
    url='https://github.com/orf/petal',
    license='MIT',
    author='Tom Forbes',
    author_email='tom@tomforb.es',
    long_description=long_description,
    long_description_content_type='text/markdown',
    python_requires='>=3.6',
    install_requires=[
        'loguru',
        'requests',
        'grpcio',
        'grpcio-tools',
        'click',
        'mypy-protobuf',
    ],
    entry_points={
        'console_scripts': [
            'petal = petal.cli:main',
            'protoc-gen-extract-streaming = petal.protoc_gen_extract_streaming:main'
        ]
    },
)
